/*
Simple non-interactive CD Player

Copyright (C) 2001,2002 Witold Filipczyk
Email: witekfl@poczta.onet.pl
This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.  See the file COPYING.

This program is distributed in hope that it will be useful,
but WITHOUT ANY WARRANTY; without even an implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/


#include <fcntl.h>
#include <linux/cdrom.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>

int drive;
struct cdrom_tochdr diskinfo;
struct cdrom_subchnl sub;

void save(pid_t pid)
{
	FILE *fd = fopen("cdplayer.pid","w");
	if (fd) {
		fprintf(fd,"%d\n", pid);
		fclose(fd);
	}
}

void killer(void)
{
	FILE *fd = fopen("cdplayer.pid","r");
	if (fd) {
		pid_t pid;
	  fscanf(fd, "%d", &pid);
	  fclose(fd);
	  unlink("cdplayer.pid");
	  kill(pid, SIGINT);
	}
}

void end(int i)
{
	killer();
	exit(0);
}

void play(__u8 f, __u8 l)
{
	struct cdrom_ti tracks;
	
	tracks.cdti_trk0 = f;
	tracks.cdti_trk1 = l;
	ioctl(drive, CDROMPLAYTRKIND, &tracks);
}

__u8 getCurrentTrack(void)
{

	ioctl(drive, CDROMSUBCHNL, &sub);
	return sub.cdsc_trk;
}

void prev(void)
{
	__u8 track;
	
	killer();
	track = getCurrentTrack();
	if (track > diskinfo.cdth_trk0) play(--track, diskinfo.cdth_trk1);
}

void next(void)
{
	__u8 track;
	
	killer();
	track = getCurrentTrack();
	if (track < diskinfo.cdth_trk1) play(++track, diskinfo.cdth_trk1);
}
	
void pauseORcont()
{
	sub.cdsc_format = CDROM_MSF;
	ioctl(drive, CDROMSUBCHNL, &sub);	
	if (sub.cdsc_audiostatus == CDROM_AUDIO_PLAY) ioctl(drive, CDROMPAUSE);
	else ioctl(drive, CDROMRESUME);
}
	

void almost_daemon(void)
{
	pid_t pid;
	
	close(0);
	close(1);
	close(2);
	killer();
	signal(SIGTERM, end);
	pid = fork();
	if (pid) {
		save(pid);
		exit(0);
	}
	setpgrp();
}	

void sleeping_loop(void)
{
	for (;;) {
			usleep(250000);
			ioctl(drive, CDROMSUBCHNL, &sub);	
			if (sub.cdsc_audiostatus == CDROM_AUDIO_COMPLETED) break;
		}
}


void rando(void)
{
	__u8 track;
	struct timeval tv;
	
	almost_daemon();
	for (;;) {
		gettimeofday(&tv, NULL);
		srandom(tv.tv_usec);
		track = random() % (diskinfo.cdth_trk1 - diskinfo.cdth_trk0 + 1)
		 + diskinfo.cdth_trk0;
		play(track, track);
		sleeping_loop();
	}		
}


void loop(void)
{
	almost_daemon();
	for (;;) {
		play(diskinfo.cdth_trk0, diskinfo.cdth_trk1);
		sleeping_loop();
	}
}

int main(int argc, char **argv)
{
	char *t;

	drive = open("/dev/cdrom", O_RDONLY | O_NONBLOCK);
	if (drive > 0) {
		chdir(getenv("TMPDIR"));
		sub.cdsc_format = CDROM_MSF;
		ioctl(drive, CDROMREADTOCHDR, &diskinfo);
		if (argc>1) {
			t = strrchr(argv[1], '=');
			if (t) {
			    switch ((char)*++t) {
				case 'F': /* First */
				    killer();
				    play(diskinfo.cdth_trk0, diskinfo.cdth_trk1); 
				    return 0;
				case 'L':
				    if (t[1] == 'o') loop(); /* Loop */
				    killer(); /* Last */
				    play(diskinfo.cdth_trk1, diskinfo.cdth_trk1);
				    return 0;
				case 'N': /* Next */
			    	    next();
				    return 0;
				case 'P': 
				    if (t[1] == 'r') prev(); /* Prev */
				    else pauseORcont(); /* Pause|Resume */
				    return 0;
				case 'R':
				    rando(); /* Random play */
				    break;
				case 'S':
				    killer(); /* Stop */
				    ioctl(drive, CDROMSTOP);
				    return 0;
				default: /* play from =NR track */
				    killer();
				    play(atoi(t), diskinfo.cdth_trk1);
				    return 0;
				}
			    }
			}
		killer(); /* play whole CD */
		play(diskinfo.cdth_trk0, diskinfo.cdth_trk1);
		return 0;
	}
	return 1;
}
